--conio.e,
--Borland-ish console functions for dos or dosbox text mode only

--summary of globals:

--function getch()          returns key without echo and with ^c break off
--procedure insline()
--procedure delline()
--procedure gotoxy(x,y)
--function wherexy()        returns cursor {x,y}
--procedure clrscr()
--procedure clreol()
--function query(qstr)      prints query string, returns answer string
--procedure save_screen()
--procedure restore_screen()

--  and the following are useful with query() --
--function atoi(str)        extracts integer from string (or returns "nan")
--function atof(str)        extracts float from string (or returns "nan")


include machine.e

constant SCREEN_ADDR = #B8000
atom SCREEN_BUFFER  SCREEN_BUFFER=0
sequence SCREEN_XY

global function getch()
  sequence reg
  reg=repeat(0,10)
  reg[REG_AX]=#0700
  reg=dos_interrupt(#21,reg)
  return and_bits(reg[REG_AX],#00FF)
  end function

global procedure insline()
  sequence reg
  reg=repeat(0,10)
  reg[REG_AX]=#0300
  reg=dos_interrupt(#10,reg)
  reg[REG_AX]=#0701
  reg[REG_BX]=#0700
  reg[REG_CX]=and_bits(reg[REG_DX],#FF00)
  reg[REG_DX]=#184F
  reg=dos_interrupt(#10,reg)
  end procedure

global procedure delline()
  sequence reg
  reg=repeat(0,10)
  reg[REG_AX]=#0300
  reg=dos_interrupt(#10,reg)
  reg[REG_AX]=#0601
  reg[REG_BX]=#0700
  reg[REG_CX]=and_bits(reg[REG_DX],#FF00)
  reg[REG_DX]=#184F
  reg=dos_interrupt(#10,reg)
  end procedure

global procedure gotoxy(integer x, integer y)
  position(y,x)
  end procedure

global function wherexy()
  sequence reg
  reg=repeat(0,10)
  reg[REG_AX]=#0300
  reg=dos_interrupt(#10,reg)
  return {1+and_bits(#00FF,reg[REG_DX]),1+and_bits(#FF00,reg[REG_DX])/#100}
  end function

global procedure clrscr()
  clear_screen()
  end procedure

global procedure clreol()
  sequence reg
  reg=repeat(0,10)
  reg[REG_AX]=#0300
  reg=dos_interrupt(#10,reg)
  reg[REG_AX]=#0600
  reg[REG_BX]=#0700
  reg[REG_CX]=reg[REG_DX]
  reg[REG_DX]=and_bits(reg[REG_DX],#FF00)+#004F
  reg=dos_interrupt(#10,reg)
  end procedure

procedure putch(integer x)  --not global, used in query()
  sequence reg
  reg=repeat(0,10)
  reg[REG_AX]=#0E00+x
  reg=dos_interrupt(#10,reg)
  end procedure

global function query(sequence qs)
  integer zz, ia
  sequence as
  puts(1,qs&" ? ")
  as=""
  ia=0
  while(1) do
    zz=getch()
    if zz>=' ' and zz<='~' then
      as=append(as,zz)
      ia=ia+1
      putch(zz)
    elsif (ia and zz=8) then  --backspace
      ia=ia-1
      as=as[1..ia]
      putch(8) putch(' ') putch(8)
    elsif zz='\r' or zz='\n' then exit
    elsif not zz then zz=getch()  --dump zero and follower
    elsif zz=27 then as="" exit  --escape forces null return
    end if
    end while
  puts(1,"\n")
  return as
  end function

global function atoi(sequence s)
  integer n, minus, c, k, issign
  sequence s1
  s1=""
  minus=0 k=0 issign=0
  for i=1 to length(s) do
    c=s[i]
    if c>='0' and c<='9' then s1=append(s1,c) k=k+1
      elsif c='-' and k=0 and issign=0 then minus=1 issign=1
      elsif c='+' and k=0 and issign=0 then issign=1
      elsif c=' ' and k=0 and issign=0 then  --ignore leading blank
      else exit  --parsing stops at unexpected character
      end if
    end for
  if k=0 then return "nan" end if
  n=0
  for i=1 to k do n=10*n+(s1[i]-'0') end for
  if minus then n= -n end if
  return n
  end function

global function atof(sequence s)
  atom x, fp
  integer c, minus, iexp, eminus,
    issign, isd, isesign, ise, ki, kf, ke
  sequence si, sf, se
  si="" sf="" se=""
  minus=0 eminus=0 issign=0 isd=0 ise=0 isesign=0
  ki=0 kf=0 ke=0
  for i=1 to length(s) do
    c=s[i]
    if c>='0' and c<='9' then
      if ise then se=append(se,c) ke=ke+1
        elsif isd then sf=append(sf,c) kf=kf+1
        else si=append(si,c) ki=ki+1
        end if
      elsif c='-' and isd=0 and ki=0 and issign=0 then minus=1 issign=1
      elsif c='+' and isd=0 and ki=0 and issign=0 then issign=1
      elsif c='-' and ise and ke=0 and isesign=0 then eminus=1 isesign=1
      elsif c='+' and ise and ke=0 and isesign=0 then isesign=1
      elsif c='.' and isd=0 then isd=1
      elsif (c='e' or c='E') and ise=0 then ise=1
      elsif c=' ' and ki=0 and isd=0 and ise=0 and issign=0 then
        --ignore leading blank
      else exit  --parsing stops at unexpected character
      end if
    end for
  if ki=0 and kf=0 then return "nan" end if
  x=0.0
  for i=1 to ki do x=10*x+(si[i]-'0') end for
  fp=1.0
  for i=1 to kf do fp=0.1*fp x=x+fp*(sf[i]-'0') end for
  if minus then x= -x end if
  iexp=0
  for i=1 to ke do iexp=10*iexp+(se[i]-'0') end for
  if eminus then iexp= -iexp end if
  x=x*power(10,iexp)
  return x
  end function

global procedure save_screen()
  if SCREEN_BUFFER=0 then SCREEN_BUFFER=allocate(4000) end if
  mem_copy(SCREEN_BUFFER,SCREEN_ADDR,4000)
  SCREEN_XY=wherexy()
  end procedure

global procedure restore_screen()
  mem_copy(SCREEN_ADDR,SCREEN_BUFFER,4000)
  gotoxy(SCREEN_XY[1],SCREEN_XY[2])
  free(SCREEN_BUFFER)
  SCREEN_BUFFER=0
  end procedure
